//
//	JdbcEncodingConverter.java
//
//	version 1.0  98/12/12
//	author  Mitsumasa IWATSUKI
//

package org.iwatsuki.jec;

import java.io.*;
import java.math.BigDecimal;
import java.sql.*;
import java.util.*;

public class JdbcEncodingConverter implements Driver {
	private final static String ASCENC = "8859_1";

	static {
		try {
			DriverManager.registerDriver(new JdbcEncodingConverter());
		} catch (SQLException e) {
			throw new ExceptionInInitializerError(e);
		}
	}

	private static String encode(String s, String enc) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
		       /* else */                        new String(s.getBytes(enc), ASCENC);
	}

	private static String decode(String s, String enc) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
		       /* else */                        new String(s.getBytes(ASCENC), enc);
	}

	private static String getEncoding(String url) {
		if (url == null)
			return null;
		if (url.length() < 5 || url.substring(0, 5).toLowerCase().equals("jdbc:"))
			return null;

		int index;
		if ((index = url.indexOf(':')) == -1)
			return null;
		return url.substring(0, index);
	}

	private static String getJdbcURL(String url) {
		if (url == null)
			return null;
		if (url.length() < 5 || url.substring(0, 5).toLowerCase().equals("jdbc:"))
			return null;

		int index;
		if ((index = url.indexOf(':')) == -1)
			return null;
		return url.substring(++index);
	}

	private static String encodeURL(String url, String enc) throws UnsupportedEncodingException {
		if (url == null)
			return null;

		int index;
		if ((index = url.indexOf(':')) == -1)
			return null;
		if ((index = url.indexOf(':', ++index)) == -1)
			return null;
		++index;
		return url.substring(0, index) + encode(url.substring(index), enc);
	}

	private static Properties encodeInfo(Properties info, String enc) throws UnsupportedEncodingException {
		if (info == null)
			return null;

		Properties encInfo = new Properties();
		Enumeration keys = info.keys();
		while (keys.hasMoreElements()) {
			Object key = keys.nextElement();
			Object value = info.get(key);
			encInfo.put((key instanceof String) ? encode((String)key, enc) : key,
			            (value instanceof String) ? encode((String)value, enc) : value);
		}
		return encInfo;
	}

	private static DriverPropertyInfo[] decodePropertyInfo(DriverPropertyInfo[] dpi, String enc) throws UnsupportedEncodingException {
		if (dpi == null)
			return null;

		DriverPropertyInfo[] decDpi = new DriverPropertyInfo[dpi.length];
		for (int i = 0; i < dpi.length; i++) {
			decDpi[i].name = decode(dpi[i].name, enc);
			decDpi[i].description = decode(dpi[i].description, enc);
			decDpi[i].value = decode(dpi[i].value, enc);
			if (dpi[i].choices != null) {
				int n = dpi[i].choices.length;
				String[] choices = new String[n];
				for (int c = 0; c < n; c++)
					choices[c] = (dpi[i].choices[c] == null) ? null : decode(dpi[i].choices[c], enc);
				decDpi[i].choices = choices;
			} else
				decDpi[i].choices = null;
		}
		return decDpi;
	}

	//----------------------------------------------------------------------

	public Connection connect(String url, Properties info) throws SQLException {
		String enc = getEncoding(url);
		if (enc == null)
			return null;

		try {
			String encUrl = encodeURL(getJdbcURL(url), enc);
			if (encUrl == null)
				return null;
			Properties encInfo = encodeInfo(info, enc);

			Driver d = DriverManager.getDriver(encUrl);
			Connection con = d.connect(encUrl, encInfo);
			return (con == null) ? null : new ConnectionWrap(con, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	public boolean acceptsURL(String url) throws SQLException {
		String enc = getEncoding(url);
		if (enc == null)
			return false;

		try {
			String encUrl = encodeURL(getJdbcURL(url), enc);
			if (encUrl == null)
				return false;

			Driver d = DriverManager.getDriver(encUrl);
			return d.acceptsURL(encUrl);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			return false;
		}
	}

	public DriverPropertyInfo[] getPropertyInfo(String url, Properties info) throws SQLException {
		DriverPropertyInfo[] emptyDpi = new DriverPropertyInfo[0];
		String enc = getEncoding(url);
		if (enc == null)
			return emptyDpi;

		try {
			String encUrl = encodeURL(getJdbcURL(url), enc);
			if (encUrl == null)
				return emptyDpi;
			Properties encInfo = encodeInfo(info, enc);

			Driver d = DriverManager.getDriver(encUrl);
			return decodePropertyInfo(d.getPropertyInfo(encUrl, encInfo), enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			return emptyDpi;
		}
	}

	public int getMajorVersion() {
		return 1;
	}

	public int getMinorVersion() {
		return 0;
	}

	public boolean jdbcCompliant() {
		return false;
	}
}

class ConnectionWrap implements Connection {
	private Connection con;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";

	public ConnectionWrap(Connection con, String enc) {
		this.con = con;
		this.enc = enc;
	}

	private String encode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
		       /* else */                        new String(s.getBytes(enc), ASCENC);
	}

	private String decode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
		       /* else */                        new String(s.getBytes(ASCENC), enc);
	}

	//----------------------------------------------------------------------

	public Statement createStatement() throws SQLException {
		try {
			Statement stmt = con.createStatement();
			return (stmt == null) ? null : new StatementWrap(stmt, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public PreparedStatement prepareStatement(String sql) throws SQLException {
		try {
			PreparedStatement pstmt = con.prepareStatement(encode(sql));
			return (pstmt == null) ? null : new PreparedStatementWrap(pstmt, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public CallableStatement prepareCall(String sql) throws SQLException {
		try {
			CallableStatement cstmt = con.prepareCall(encode(sql));
			return (cstmt == null) ? null : new CallableStatementWrap(cstmt, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String nativeSQL(String sql) throws SQLException {
		try {
			return decode(con.nativeSQL(encode(sql)));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public void setAutoCommit(boolean autoCommit) throws SQLException {
		try {
			con.setAutoCommit(autoCommit);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean getAutoCommit() throws SQLException {
		try {
			return con.getAutoCommit();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void commit() throws SQLException {
		try {
			con.commit();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void rollback() throws SQLException {
		try {
			con.rollback();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void close() throws SQLException {
		try {
			con.close();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isClosed() throws SQLException {
		try {
			return con.isClosed();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//======================================================================
	// Advanced features:

	public DatabaseMetaData getMetaData() throws SQLException {
		try {
			DatabaseMetaData dm = con.getMetaData();
			return (dm == null) ? null : new DatabaseMetaDataWrap(dm, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setReadOnly(boolean readOnly) throws SQLException {
		try {
			con.setReadOnly(readOnly);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isReadOnly() throws SQLException {
		try {
			return con.isReadOnly();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setCatalog(String catalog) throws SQLException {
		try {
			con.setCatalog(encode(catalog));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getCatalog() throws SQLException {
		try {
			return decode(con.getCatalog());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public void setTransactionIsolation(int level) throws SQLException {
		try {
			con.setTransactionIsolation(level);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getTransactionIsolation() throws SQLException {
		try {
			return con.getTransactionIsolation();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public SQLWarning getWarnings() throws SQLException {
		try {
			SQLWarning w = con.getWarnings();
			return (w == null) ? null : new SQLWarningWrap(w, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void clearWarnings() throws SQLException {
		try {
			con.clearWarnings();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}
}

class StatementWrap implements Statement {
	private Statement stmt;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";

	public StatementWrap(Statement stmt, String enc) {
		this.stmt = stmt;
		this.enc = enc;
	}

	private String encode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
		       /* else */                        new String(s.getBytes(enc), ASCENC);
	}

	//----------------------------------------------------------------------

	public ResultSet executeQuery(String sql) throws SQLException {
		try {
			ResultSet rs = stmt.executeQuery(encode(sql));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public int executeUpdate(String sql) throws SQLException {
		try {
			return stmt.executeUpdate(encode(sql));
		} catch (SQLException e) {
			if (e instanceof DataTruncation)
				throw new DataTruncationWrap((DataTruncation)e, enc);
			else
				throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public void close() throws SQLException {
		try {
			stmt.close();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//----------------------------------------------------------------------

	public int getMaxFieldSize() throws SQLException {
		try {
			return stmt.getMaxFieldSize();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setMaxFieldSize(int max) throws SQLException {
		try {
			stmt.setMaxFieldSize(max);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxRows() throws SQLException {
		try {
			return stmt.getMaxRows();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setMaxRows(int max) throws SQLException {
		try {
			stmt.setMaxRows(max);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setEscapeProcessing(boolean enable) throws SQLException {
		try {
			stmt.setEscapeProcessing(enable);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getQueryTimeout() throws SQLException {
		try {
			return stmt.getQueryTimeout();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setQueryTimeout(int seconds) throws SQLException {
		try {
			stmt.setQueryTimeout(seconds);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void cancel() throws SQLException {
		try {
			stmt.cancel();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public SQLWarning getWarnings() throws SQLException {
		try {
			SQLWarning w = stmt.getWarnings();
			if (w == null)
				return null;
			else if (w instanceof DataTruncation)
				return new DataTruncationWrap((DataTruncation)w, enc);
			else
				return new SQLWarningWrap(w, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void clearWarnings() throws SQLException {
		try {
			stmt.clearWarnings();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setCursorName(String name) throws SQLException {
		try {
			stmt.setCursorName(encode(name));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	//----------------------- Multiple Results --------------------------

	public boolean execute(String sql) throws SQLException {
		try {
			return stmt.execute(encode(sql));
		} catch (SQLException e) {
			if (e instanceof DataTruncation)
				throw new DataTruncationWrap((DataTruncation)e, enc);
			else
				throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getResultSet() throws SQLException {
		try {
			ResultSet rs = stmt.getResultSet();
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getUpdateCount() throws SQLException {
		try {
			return stmt.getUpdateCount();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean getMoreResults() throws SQLException {
		try {
			return stmt.getMoreResults();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}
}

class PreparedStatementWrap extends StatementWrap implements PreparedStatement {
	private PreparedStatement pstmt;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";
	private static final String UNEXPECTED_ERROR_STATE = "58033";

	public PreparedStatementWrap(PreparedStatement pstmt, String enc) {
		super(pstmt, enc);
		this.pstmt = pstmt;
		this.enc = enc;
	}

	private String encode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
		       /* else */                        new String(s.getBytes(enc), ASCENC);
	}

	//----------------------------------------------------------------------

	public ResultSet executeQuery() throws SQLException {
		try {
			ResultSet rs = pstmt.executeQuery();
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int executeUpdate() throws SQLException {
		try {
			return pstmt.executeUpdate();
		} catch (SQLException e) {
			if (e instanceof DataTruncation)
				throw new DataTruncationWrap((DataTruncation)e, enc);
			else
				throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setNull(int parameterIndex, int sqlType) throws SQLException {
		try {
			pstmt.setNull(parameterIndex, sqlType);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setBoolean(int parameterIndex, boolean x) throws SQLException {
		try {
			pstmt.setBoolean(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setByte(int parameterIndex, byte x) throws SQLException {
		try {
			pstmt.setByte(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setShort(int parameterIndex, short x) throws SQLException {
		try {
			pstmt.setShort(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setInt(int parameterIndex, int x) throws SQLException {
		try {
			pstmt.setInt(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setLong(int parameterIndex, long x) throws SQLException {
		try {
			pstmt.setLong(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setFloat(int parameterIndex, float x) throws SQLException {
		try {
			pstmt.setFloat(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setDouble(int parameterIndex, double x) throws SQLException {
		try {
			pstmt.setDouble(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException {
		try {
			pstmt.setBigDecimal(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setString(int parameterIndex, String x) throws SQLException {
		try {
			pstmt.setString(parameterIndex, encode(x));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public void setBytes(int parameterIndex, byte x[]) throws SQLException {
		try {
			pstmt.setBytes(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setDate(int parameterIndex, java.sql.Date x) throws SQLException {
		try {
			pstmt.setDate(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setTime(int parameterIndex, java.sql.Time x) throws SQLException {
		try {
			pstmt.setTime(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setTimestamp(int parameterIndex, java.sql.Timestamp x) throws SQLException {
		try {
			pstmt.setTimestamp(parameterIndex, x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setAsciiStream(int parameterIndex, java.io.InputStream x, int length) throws SQLException {
		try {
			if (x != null)
				x = new InputStreamWrap(x, enc, true);
			pstmt.setAsciiStream(parameterIndex, x, length);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void setUnicodeStream(int parameterIndex, java.io.InputStream x, int length) throws SQLException {
		try {
			if (x != null) {
				UnicodeToBinaryStreamWrap in = new UnicodeToBinaryStreamWrap(x, length, enc);
				x = new InputStreamWrap(in, enc, true);
				length = in.getLength();
			}
			pstmt.setBinaryStream(parameterIndex, x, length);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (IOException e) {
			throw new SQLException(e.getMessage(), UNEXPECTED_ERROR_STATE);
		}
	}

	public void setBinaryStream(int parameterIndex, java.io.InputStream x, int length) throws SQLException {
		try {
			if (x != null)
				x = new InputStreamWrap(x, enc, true);
			pstmt.setBinaryStream(parameterIndex, x, length);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void clearParameters() throws SQLException {
		try {
			pstmt.clearParameters();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//----------------------------------------------------------------------
	// Advanced features:

	public void setObject(int parameterIndex, Object x, int targetSqlType, int scale) throws SQLException {
		try {
			pstmt.setObject(parameterIndex, (x instanceof String) ? encode((String)x) : x, targetSqlType, scale);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public void setObject(int parameterIndex, Object x, int targetSqlType) throws SQLException {
		try {
			pstmt.setObject(parameterIndex, (x instanceof String) ? encode((String)x) : x, targetSqlType);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public void setObject(int parameterIndex, Object x) throws SQLException {
		try {
			pstmt.setObject(parameterIndex, (x instanceof String) ? encode((String)x) : x);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean execute() throws SQLException {
		try {
			return pstmt.execute();
		} catch (SQLException e) {
			if (e instanceof DataTruncation)
				throw new DataTruncationWrap((DataTruncation)e, enc);
			else
				throw new SQLExceptionWrap(e, enc);
		}
	}
}

class CallableStatementWrap extends PreparedStatementWrap implements CallableStatement {
	private CallableStatement cstmt;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";

	public CallableStatementWrap(CallableStatement cstmt, String enc) {
		super(cstmt, enc);
		this.cstmt = cstmt;
		this.enc = enc;
	}

	private String decode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
		       /* else */                        new String(s.getBytes(ASCENC), enc);
	}

	//----------------------------------------------------------------------

	public void registerOutParameter(int parameterIndex, int sqlType) throws SQLException {
		try {
			cstmt.registerOutParameter(parameterIndex, sqlType);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void registerOutParameter(int parameterIndex, int sqlType, int scale) throws SQLException {
		try {
			cstmt.registerOutParameter(parameterIndex, sqlType, scale);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean wasNull() throws SQLException {
		try {
			return cstmt.wasNull();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getString(int parameterIndex) throws SQLException {
		try {
			return decode(cstmt.getString(parameterIndex));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean getBoolean(int parameterIndex) throws SQLException {
		try {
			return cstmt.getBoolean(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public byte getByte(int parameterIndex) throws SQLException {
		try {
			return cstmt.getByte(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public short getShort(int parameterIndex) throws SQLException {
		try {
			return cstmt.getShort(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getInt(int parameterIndex) throws SQLException {
		try {
			return cstmt.getInt(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public long getLong(int parameterIndex) throws SQLException {
		try {
			return cstmt.getLong(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public float getFloat(int parameterIndex) throws SQLException {
		try {
			return cstmt.getFloat(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public double getDouble(int parameterIndex) throws SQLException {
		try {
			return cstmt.getDouble(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public BigDecimal getBigDecimal(int parameterIndex, int scale) throws SQLException {
		try {
			return cstmt.getBigDecimal(parameterIndex, scale);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public byte[] getBytes(int parameterIndex) throws SQLException {
		try {
			return cstmt.getBytes(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.sql.Date getDate(int parameterIndex) throws SQLException {
		try {
			return cstmt.getDate(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.sql.Time getTime(int parameterIndex) throws SQLException {
		try {
			return cstmt.getTime(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.sql.Timestamp getTimestamp(int parameterIndex) throws SQLException {
		try {
			return cstmt.getTimestamp(parameterIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//----------------------------------------------------------------------
	// Advanced features:

	public Object getObject(int parameterIndex) throws SQLException {
		try {
			Object x = cstmt.getObject(parameterIndex);
			return (x instanceof String) ? decode((String)x) : x;
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}
}

class ResultSetWrap implements ResultSet {
	private ResultSet rs;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";

	public ResultSetWrap(ResultSet rs, String enc) {
		this.rs = rs;
		this.enc = enc;
	}

	private String encode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
		       /* else */                        new String(s.getBytes(enc), ASCENC);
	}

	private String decode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
		       /* else */                        new String(s.getBytes(ASCENC), enc);
	}

	//----------------------------------------------------------------------

	public boolean next() throws SQLException {
		try {
			return rs.next();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void close() throws SQLException {
		try {
			rs.close();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean wasNull() throws SQLException {
		try {
			return rs.wasNull();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//======================================================================
	// Methods for accessing results by column index
	//======================================================================

	public String getString(int columnIndex) throws SQLException {
		try {
			return decode(rs.getString(columnIndex));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean getBoolean(int columnIndex) throws SQLException {
		try {
			return rs.getBoolean(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public byte getByte(int columnIndex) throws SQLException {
		try {
			return rs.getByte(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public short getShort(int columnIndex) throws SQLException {
		try {
			return rs.getShort(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getInt(int columnIndex) throws SQLException {
		try {
			return rs.getInt(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public long getLong(int columnIndex) throws SQLException {
		try {
			return rs.getLong(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public float getFloat(int columnIndex) throws SQLException {
		try {
			return rs.getFloat(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public double getDouble(int columnIndex) throws SQLException {
		try {
			return rs.getDouble(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public BigDecimal getBigDecimal(int columnIndex, int scale) throws SQLException {
		try {
			return rs.getBigDecimal(columnIndex, scale);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public byte[] getBytes(int columnIndex) throws SQLException {
		try {
			return rs.getBytes(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.sql.Date getDate(int columnIndex) throws SQLException {
		try {
			return rs.getDate(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.sql.Time getTime(int columnIndex) throws SQLException {
		try {
			return rs.getTime(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.sql.Timestamp getTimestamp(int columnIndex) throws SQLException {
		try {
			return rs.getTimestamp(columnIndex);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.io.InputStream getAsciiStream(int columnIndex) throws SQLException {
		try {
			InputStream x = rs.getAsciiStream(columnIndex);
			return (x == null) ? null : new InputStreamWrap(x, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public java.io.InputStream getUnicodeStream(int columnIndex) throws SQLException {
		try {
			InputStream x = rs.getBinaryStream(columnIndex);
			return (x == null) ? null : new BinaryToUnicodeStreamWrap(new InputStreamWrap(x, enc), enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.io.InputStream getBinaryStream(int columnIndex) throws SQLException {
		try {
			InputStream x = rs.getBinaryStream(columnIndex);
			return (x == null) ? null : new InputStreamWrap(x, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//======================================================================
	// Methods for accessing results by column name
	//======================================================================

	public String getString(String columnName) throws SQLException {
		try {
			return decode(rs.getString(encode(columnName)));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean getBoolean(String columnName) throws SQLException {
		try {
			return rs.getBoolean(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public byte getByte(String columnName) throws SQLException {
		try {
			return rs.getByte(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public short getShort(String columnName) throws SQLException {
		try {
			return rs.getShort(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public int getInt(String columnName) throws SQLException {
		try {
			return rs.getInt(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public long getLong(String columnName) throws SQLException {
		try {
			return rs.getLong(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public float getFloat(String columnName) throws SQLException {
		try {
			return rs.getFloat(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public double getDouble(String columnName) throws SQLException {
		try {
			return rs.getDouble(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public BigDecimal getBigDecimal(String columnName, int scale) throws SQLException {
		try {
			return rs.getBigDecimal(encode(columnName), scale);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public byte[] getBytes(String columnName) throws SQLException {
		try {
			return rs.getBytes(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.sql.Date getDate(String columnName) throws SQLException {
		try {
			return rs.getDate(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.sql.Time getTime(String columnName) throws SQLException {
		try {
			return rs.getTime(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.sql.Timestamp getTimestamp(String columnName) throws SQLException {
		try {
			return rs.getTimestamp(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.io.InputStream getAsciiStream(String columnName) throws SQLException {
		try {
			InputStream x = rs.getAsciiStream(encode(columnName));
			return (x == null) ? null : new InputStreamWrap(x, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.io.InputStream getUnicodeStream(String columnName) throws SQLException {
		try {
			InputStream x = rs.getBinaryStream(encode(columnName));
			return (x == null) ? null : new BinaryToUnicodeStreamWrap(new InputStreamWrap(x, enc), enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public java.io.InputStream getBinaryStream(String columnName) throws SQLException {
		try {
			InputStream x = rs.getBinaryStream(encode(columnName));
			return (x == null) ? null : new InputStreamWrap(x, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	//=====================================================================
	// Advanced features:
	//=====================================================================

	public SQLWarning getWarnings() throws SQLException {
		try {
			SQLWarning w = rs.getWarnings();
			if (w == null)
				return null;
			else if (w instanceof DataTruncation)
				return new DataTruncationWrap((DataTruncation)w, enc);
			else
				return new SQLWarningWrap(w, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public void clearWarnings() throws SQLException {
		try {
			rs.clearWarnings();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getCursorName() throws SQLException {
		try {
			return decode(rs.getCursorName());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSetMetaData getMetaData() throws SQLException {
		try {
			ResultSetMetaData rsmd = rs.getMetaData();
			return (rsmd == null) ? null : new ResultSetMetaDataWrap(rsmd, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public Object getObject(int columnIndex) throws SQLException {
		try {
			Object x = rs.getObject(columnIndex);
			return (x instanceof String) ? decode((String)x) : x;
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public Object getObject(String columnName) throws SQLException {
		try {
			Object x = rs.getObject(encode(columnName));
			return (x instanceof String) ? decode((String)x) : x;
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	//----------------------------------------------------------------

	public int findColumn(String columnName) throws SQLException {
		try {
			return rs.findColumn(encode(columnName));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}
}

class DatabaseMetaDataWrap implements DatabaseMetaData {
	private DatabaseMetaData dm;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";

	public DatabaseMetaDataWrap(DatabaseMetaData dm, String enc) {
		this.dm = dm;
		this.enc = enc;
	}

	private String encode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
		       /* else */                        new String(s.getBytes(enc), ASCENC);
	}

	private String decode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
		       /* else */                        new String(s.getBytes(ASCENC), enc);
	}

	//----------------------------------------------------------------------
	// First, a variety of minor information about the target database.

	public boolean allProceduresAreCallable() throws SQLException {
		try {
			return dm.allProceduresAreCallable();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean allTablesAreSelectable() throws SQLException {
		try {
			return dm.allTablesAreSelectable();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getURL() throws SQLException {
		try {
			return decode(dm.getURL());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getUserName() throws SQLException {
		try {
			return decode(dm.getUserName());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean isReadOnly() throws SQLException {
		try {
			return dm.isReadOnly();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean nullsAreSortedHigh() throws SQLException {
		try {
			return dm.nullsAreSortedHigh();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean nullsAreSortedLow() throws SQLException {
		try {
			return dm.nullsAreSortedLow();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean nullsAreSortedAtStart() throws SQLException {
		try {
			return dm.nullsAreSortedAtStart();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean nullsAreSortedAtEnd() throws SQLException {
		try {
			return dm.nullsAreSortedAtEnd();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getDatabaseProductName() throws SQLException {
		try {
			return decode(dm.getDatabaseProductName());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getDatabaseProductVersion() throws SQLException {
		try {
			return decode(dm.getDatabaseProductVersion());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getDriverName() throws SQLException {
		try {
			return decode(dm.getDriverName());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getDriverVersion() throws SQLException {
		try {
			return decode(dm.getDriverVersion());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public int getDriverMajorVersion() {
		return dm.getDriverMajorVersion();
	}

	public int getDriverMinorVersion() {
		return dm.getDriverMinorVersion();
	}

	public boolean usesLocalFiles() throws SQLException {
		try {
			return dm.usesLocalFiles();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean usesLocalFilePerTable() throws SQLException {
		try {
			return dm.usesLocalFilePerTable();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsMixedCaseIdentifiers() throws SQLException {
		try {
			return dm.supportsMixedCaseIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean storesUpperCaseIdentifiers() throws SQLException {
		try {
			return dm.storesUpperCaseIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean storesLowerCaseIdentifiers() throws SQLException {
		try {
			return dm.storesLowerCaseIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean storesMixedCaseIdentifiers() throws SQLException {
		try {
			return dm.storesMixedCaseIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsMixedCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.supportsMixedCaseQuotedIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean storesUpperCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.storesUpperCaseQuotedIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean storesLowerCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.storesLowerCaseQuotedIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean storesMixedCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.storesMixedCaseQuotedIdentifiers();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getIdentifierQuoteString() throws SQLException {
		try {
			return decode(dm.getIdentifierQuoteString());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getSQLKeywords() throws SQLException {
		try {
			return decode(dm.getSQLKeywords());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getNumericFunctions() throws SQLException {
		try {
			return decode(dm.getNumericFunctions());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getStringFunctions() throws SQLException {
		try {
			return decode(dm.getStringFunctions());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getSystemFunctions() throws SQLException {
		try {
			return decode(dm.getSystemFunctions());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getTimeDateFunctions() throws SQLException {
		try {
			return decode(dm.getTimeDateFunctions());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getSearchStringEscape() throws SQLException {
		try {
			return decode(dm.getSearchStringEscape());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getExtraNameCharacters() throws SQLException {
		try {
			return decode(dm.getExtraNameCharacters());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	//--------------------------------------------------------------------
	// Functions describing which features are supported.

	public boolean supportsAlterTableWithAddColumn() throws SQLException {
		try {
			return dm.supportsAlterTableWithAddColumn();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsAlterTableWithDropColumn() throws SQLException {
		try {
			return dm.supportsAlterTableWithDropColumn();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsColumnAliasing() throws SQLException {
		try {
			return dm.supportsColumnAliasing();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean nullPlusNonNullIsNull() throws SQLException {
		try {
			return dm.nullPlusNonNullIsNull();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsConvert() throws SQLException {
		try {
			return dm.supportsConvert();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsConvert(int fromType, int toType) throws SQLException {
		try {
			return dm.supportsConvert(fromType, toType);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsTableCorrelationNames() throws SQLException {
		try {
			return dm.supportsTableCorrelationNames();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsDifferentTableCorrelationNames() throws SQLException {
		try {
			return dm.supportsDifferentTableCorrelationNames();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsExpressionsInOrderBy() throws SQLException {
		try {
			return dm.supportsExpressionsInOrderBy();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsOrderByUnrelated() throws SQLException {
		try {
			return dm.supportsOrderByUnrelated();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsGroupBy() throws SQLException {
		try {
			return dm.supportsGroupBy();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsGroupByUnrelated() throws SQLException {
		try {
			return dm.supportsGroupByUnrelated();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsGroupByBeyondSelect() throws SQLException {
		try {
			return dm.supportsGroupByBeyondSelect();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsLikeEscapeClause() throws SQLException {
		try {
			return dm.supportsLikeEscapeClause();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsMultipleResultSets() throws SQLException {
		try {
			return dm.supportsMultipleResultSets();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsMultipleTransactions() throws SQLException {
		try {
			return dm.supportsMultipleTransactions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsNonNullableColumns() throws SQLException {
		try {
			return dm.supportsNonNullableColumns();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsMinimumSQLGrammar() throws SQLException {
		try {
			return dm.supportsMinimumSQLGrammar();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCoreSQLGrammar() throws SQLException {
		try {
			return dm.supportsCoreSQLGrammar();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsExtendedSQLGrammar() throws SQLException {
		try {
			return dm.supportsExtendedSQLGrammar();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsANSI92EntryLevelSQL() throws SQLException {
		try {
			return dm.supportsANSI92EntryLevelSQL();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsANSI92IntermediateSQL() throws SQLException {
		try {
			return dm.supportsANSI92IntermediateSQL();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsANSI92FullSQL() throws SQLException {
		try {
			return dm.supportsANSI92FullSQL();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsIntegrityEnhancementFacility() throws SQLException {
		try {
			return dm.supportsIntegrityEnhancementFacility();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsOuterJoins() throws SQLException {
		try {
			return dm.supportsOuterJoins();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsFullOuterJoins() throws SQLException {
		try {
			return dm.supportsFullOuterJoins();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsLimitedOuterJoins() throws SQLException {
		try {
			return dm.supportsLimitedOuterJoins();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getSchemaTerm() throws SQLException {
		try {
			return decode(dm.getSchemaTerm());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getProcedureTerm() throws SQLException {
		try {
			return decode(dm.getProcedureTerm());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getCatalogTerm() throws SQLException {
		try {
			return decode(dm.getCatalogTerm());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean isCatalogAtStart() throws SQLException {
		try {
			return dm.isCatalogAtStart();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getCatalogSeparator() throws SQLException {
		try {
			return decode(dm.getCatalogSeparator());
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean supportsSchemasInDataManipulation() throws SQLException {
		try {
			return dm.supportsSchemasInDataManipulation();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSchemasInProcedureCalls() throws SQLException {
		try {
			return dm.supportsSchemasInProcedureCalls();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSchemasInTableDefinitions() throws SQLException {
		try {
			return dm.supportsSchemasInTableDefinitions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSchemasInIndexDefinitions() throws SQLException {
		try {
			return dm.supportsSchemasInIndexDefinitions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSchemasInPrivilegeDefinitions() throws SQLException {
		try {
			return dm.supportsSchemasInPrivilegeDefinitions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCatalogsInDataManipulation() throws SQLException {
		try {
			return dm.supportsCatalogsInDataManipulation();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCatalogsInProcedureCalls() throws SQLException {
		try {
			return dm.supportsCatalogsInProcedureCalls();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCatalogsInTableDefinitions() throws SQLException {
		try {
			return dm.supportsCatalogsInTableDefinitions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCatalogsInIndexDefinitions() throws SQLException {
		try {
			return dm.supportsCatalogsInIndexDefinitions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCatalogsInPrivilegeDefinitions() throws SQLException {
		try {
			return dm.supportsCatalogsInPrivilegeDefinitions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsPositionedDelete() throws SQLException {
		try {
			return dm.supportsPositionedDelete();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsPositionedUpdate() throws SQLException {
		try {
			return dm.supportsPositionedUpdate();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSelectForUpdate() throws SQLException {
		try {
			return dm.supportsSelectForUpdate();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsStoredProcedures() throws SQLException {
		try {
			return dm.supportsStoredProcedures();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSubqueriesInComparisons() throws SQLException {
		try {
			return dm.supportsSubqueriesInComparisons();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSubqueriesInExists() throws SQLException {
		try {
			return dm.supportsSubqueriesInExists();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSubqueriesInIns() throws SQLException {
		try {
			return dm.supportsSubqueriesInIns();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsSubqueriesInQuantifieds() throws SQLException {
		try {
			return dm.supportsSubqueriesInQuantifieds();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsCorrelatedSubqueries() throws SQLException {
		try {
			return dm.supportsCorrelatedSubqueries();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsUnion() throws SQLException {
		try {
			return dm.supportsUnion();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsUnionAll() throws SQLException {
		try {
			return dm.supportsUnionAll();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsOpenCursorsAcrossCommit() throws SQLException {
		try {
			return dm.supportsOpenCursorsAcrossCommit();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsOpenCursorsAcrossRollback() throws SQLException {
		try {
			return dm.supportsOpenCursorsAcrossRollback();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsOpenStatementsAcrossCommit() throws SQLException {
		try {
			return dm.supportsOpenStatementsAcrossCommit();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsOpenStatementsAcrossRollback() throws SQLException {
		try {
			return dm.supportsOpenStatementsAcrossRollback();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//----------------------------------------------------------------------
	// The following group of methods exposes various limitations
	// based on the target database with the current driver.
	// Unless otherwise specified, a result of zero means there is no
	// limit, or the limit is not known.

	public int getMaxBinaryLiteralLength() throws SQLException {
		try {
			return dm.getMaxBinaryLiteralLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxCharLiteralLength() throws SQLException {
		try {
			return dm.getMaxCharLiteralLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxColumnNameLength() throws SQLException {
		try {
			return dm.getMaxColumnNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxColumnsInGroupBy() throws SQLException {
		try {
			return dm.getMaxColumnsInGroupBy();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxColumnsInIndex() throws SQLException {
		try {
			return dm.getMaxColumnsInIndex();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxColumnsInOrderBy() throws SQLException {
		try {
			return dm.getMaxColumnsInOrderBy();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxColumnsInSelect() throws SQLException {
		try {
			return dm.getMaxColumnsInSelect();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxColumnsInTable() throws SQLException {
		try {
			return dm.getMaxColumnsInTable();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxConnections() throws SQLException {
		try {
			return dm.getMaxConnections();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxCursorNameLength() throws SQLException {
		try {
			return dm.getMaxCursorNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxIndexLength() throws SQLException {
		try {
			return dm.getMaxIndexLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxSchemaNameLength() throws SQLException {
		try {
			return dm.getMaxSchemaNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxProcedureNameLength() throws SQLException {
		try {
			return dm.getMaxProcedureNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxCatalogNameLength() throws SQLException {
		try {
			return dm.getMaxCatalogNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxRowSize() throws SQLException {
		try {
			return dm.getMaxRowSize();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean doesMaxRowSizeIncludeBlobs() throws SQLException {
		try {
			return dm.doesMaxRowSizeIncludeBlobs();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxStatementLength() throws SQLException {
		try {
			return dm.getMaxStatementLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxStatements() throws SQLException {
		try {
			return dm.getMaxStatements();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxTableNameLength() throws SQLException {
		try {
			return dm.getMaxTableNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxTablesInSelect() throws SQLException {
		try {
			return dm.getMaxTablesInSelect();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getMaxUserNameLength() throws SQLException {
		try {
			return dm.getMaxUserNameLength();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	//----------------------------------------------------------------------

	public int getDefaultTransactionIsolation() throws SQLException {
		try {
			return dm.getDefaultTransactionIsolation();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsTransactions() throws SQLException {
		try {
			return dm.supportsTransactions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsTransactionIsolationLevel(int level) throws SQLException {
		try {
			return dm.supportsTransactionIsolationLevel(level);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsDataDefinitionAndDataManipulationTransactions() throws SQLException {
		try {
			return dm.supportsDataDefinitionAndDataManipulationTransactions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean supportsDataManipulationTransactionsOnly() throws SQLException {
		try {
			return dm.supportsDataManipulationTransactionsOnly();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean dataDefinitionCausesTransactionCommit() throws SQLException {
		try {
			return dm.dataDefinitionCausesTransactionCommit();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean dataDefinitionIgnoredInTransactions() throws SQLException {
		try {
			return dm.dataDefinitionIgnoredInTransactions();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public ResultSet getProcedures(String catalog, String schemaPattern, String procedureNamePattern) throws SQLException {
		try {
			ResultSet rs = dm.getProcedures(encode(catalog), encode(schemaPattern), encode(procedureNamePattern));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getProcedureColumns(String catalog, String schemaPattern, String procedureNamePattern, String columnNamePattern) throws SQLException {
		try {
			ResultSet rs = dm.getProcedureColumns(encode(catalog), encode(schemaPattern), encode(procedureNamePattern), encode(columnNamePattern));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getTables(String catalog, String schemaPattern, String tableNamePattern, String types[]) throws SQLException {
		try {
			String encTypes[];
			if (types == null) {
				encTypes = null;
			} else {
				encTypes = new String[types.length];
				for (int i = 0; i < types.length; i++)
					encTypes[i] = encode(types[i]);
			}

			ResultSet rs = dm.getTables(encode(catalog), encode(schemaPattern), encode(tableNamePattern), encTypes);
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getSchemas() throws SQLException {
		try {
			ResultSet rs = dm.getSchemas();
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public ResultSet getCatalogs() throws SQLException {
		try {
			ResultSet rs = dm.getCatalogs();
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public ResultSet getTableTypes() throws SQLException {
		try {
			ResultSet rs = dm.getTableTypes();
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public ResultSet getColumns(String catalog, String schemaPattern, String tableNamePattern, String columnNamePattern) throws SQLException {
		try {
			ResultSet rs = dm.getColumns(encode(catalog), encode(schemaPattern), encode(tableNamePattern), encode(columnNamePattern));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getColumnPrivileges(String catalog, String schema, String table, String columnNamePattern) throws SQLException {
		try {
			ResultSet rs = dm.getColumnPrivileges(encode(catalog), encode(schema), encode(table), encode(columnNamePattern));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getTablePrivileges(String catalog, String schemaPattern, String tableNamePattern) throws SQLException {
		try {
			ResultSet rs = dm.getTablePrivileges(encode(catalog), encode(schemaPattern), encode(tableNamePattern));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getBestRowIdentifier(String catalog, String schema, String table, int scope, boolean nullable) throws SQLException {
		try {
			ResultSet rs = dm.getBestRowIdentifier(encode(catalog), encode(schema), encode(table), scope, nullable);
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getVersionColumns(String catalog, String schema, String table) throws SQLException {
		try {
			ResultSet rs = dm.getVersionColumns(encode(catalog), encode(schema), encode(table));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getPrimaryKeys(String catalog, String schema, String table) throws SQLException {
		try {
			ResultSet rs = dm.getPrimaryKeys(encode(catalog), encode(schema), encode(table));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getImportedKeys(String catalog, String schema, String table) throws SQLException {
		try {
			ResultSet rs = dm.getImportedKeys(encode(catalog), encode(schema), encode(table));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getExportedKeys(String catalog, String schema, String table) throws SQLException {
		try {
			ResultSet rs = dm.getExportedKeys(encode(catalog), encode(schema), encode(table));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getCrossReference(String primaryCatalog, String primarySchema, String primaryTable, String foreignCatalog, String foreignSchema, String foreignTable) throws SQLException {
		try {
			ResultSet rs = dm.getCrossReference(encode(primaryCatalog), encode(primarySchema), encode(primaryTable), encode(foreignCatalog), encode(foreignSchema), encode(foreignTable));
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public ResultSet getTypeInfo() throws SQLException {
		try {
			ResultSet rs = dm.getTypeInfo();
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public ResultSet getIndexInfo(String catalog, String schema, String table, boolean unique, boolean approximate) throws SQLException {
		try {
			ResultSet rs = dm.getIndexInfo(encode(catalog), encode(schema), encode(table), unique, approximate);
			return (rs == null) ? null : new ResultSetWrap(rs, enc);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}
}

class ResultSetMetaDataWrap implements ResultSetMetaData {
	private ResultSetMetaData rsmd;
	private String enc;
	private final static String ASCENC = "8859_1";
	private static final String INVALID_ENCODING_STATE = "2C000";

	public ResultSetMetaDataWrap(ResultSetMetaData rsmd, String enc) {
		this.rsmd = rsmd;
		this.enc = enc;
	}

	private String decode(String s) throws UnsupportedEncodingException {
		return (s == null)                     ? null :
		       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
		       /* else */                        new String(s.getBytes(ASCENC), enc);
	}

	//----------------------------------------------------------------------

	public int getColumnCount() throws SQLException {
		try {
			return rsmd.getColumnCount();
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isAutoIncrement(int column) throws SQLException {
		try {
			return rsmd.isAutoIncrement(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isCaseSensitive(int column) throws SQLException {
		try {
			return rsmd.isCaseSensitive(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isSearchable(int column) throws SQLException {
		try {
			return rsmd.isSearchable(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isCurrency(int column) throws SQLException {
		try {
			return rsmd.isCurrency(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int isNullable(int column) throws SQLException {
		try {
			return rsmd.isNullable(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isSigned(int column) throws SQLException {
		try {
			return rsmd.isSigned(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getColumnDisplaySize(int column) throws SQLException {
		try {
			return rsmd.getColumnDisplaySize(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getColumnLabel(int column) throws SQLException {
		try {
			return decode(rsmd.getColumnLabel(column));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getColumnName(int column) throws SQLException {
		try {
			return decode(rsmd.getColumnName(column));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getSchemaName(int column) throws SQLException {
		try {
			return decode(rsmd.getSchemaName(column));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public int getPrecision(int column) throws SQLException {
		try {
			return rsmd.getPrecision(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public int getScale(int column) throws SQLException {
		try {
			return rsmd.getScale(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getTableName(int column) throws SQLException {
		try {
			return decode(rsmd.getTableName(column));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public String getCatalogName(int column) throws SQLException {
		try {
			return decode(rsmd.getCatalogName(column));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public int getColumnType(int column) throws SQLException {
		try {
			return rsmd.getColumnType(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public String getColumnTypeName(int column) throws SQLException {
		try {
			return decode(rsmd.getColumnTypeName(column));
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		} catch (UnsupportedEncodingException e) {
			throw new SQLException(e.getMessage(), INVALID_ENCODING_STATE);
		}
	}

	public boolean isReadOnly(int column) throws SQLException {
		try {
			return rsmd.isReadOnly(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isWritable(int column) throws SQLException {
		try {
			return rsmd.isWritable(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}

	public boolean isDefinitelyWritable(int column) throws SQLException {
		try {
			return rsmd.isDefinitelyWritable(column);
		} catch (SQLException e) {
			throw new SQLExceptionWrap(e, enc);
		}
	}
}

class UnicodeToBinaryStreamWrap extends InputStream {
	private byte[] bb;
	private ByteArrayInputStream in;

	public UnicodeToBinaryStreamWrap(InputStream in, int length, String enc) throws IOException {
		int len = length / 2;
		char[] cbuf = new char[len];

		int b;
		int i;
		for (i = 0; i < len; i++) {
			b = in.read();
			if (b == -1)
				break;
			cbuf[i] = (char)(b * 256);

			b = in.read();
			if (b == -1)
				break;
			cbuf[i] += (char)b;
		}
		len = i;

		ByteArrayOutputStream out = new ByteArrayOutputStream();
		OutputStreamWriter w = new OutputStreamWriter(out, enc);
		w.write(cbuf, 0, len);
		w.flush();
		bb = out.toByteArray();
		this.in = new ByteArrayInputStream(bb);
	}

	public int getLength() {
		return bb.length;
	}

	//----------------------------------------------------------------------

	public int read() throws IOException {
		return in.read();
	}

	public void close() throws IOException {
		in.close();
	}
}

class BinaryToUnicodeStreamWrap extends InputStream {
	private InputStreamReader r;
	private int bb = -1;

	public BinaryToUnicodeStreamWrap(InputStream in, String enc) throws UnsupportedEncodingException {
		r = new InputStreamReader(in, enc);
	}

	//----------------------------------------------------------------------

	public synchronized int read() throws IOException {
		int b;
		if (bb == -1) {
			bb = r.read();
			if (bb == -1)
				return bb;
			else {
				b = bb / 256;
				bb %= 256;
				return b;	
			}
		} else {
			b = bb;
			bb = -1;
			return b;	
		}
	}

	public void close() throws IOException {
		r.close();
	}
}

class InputStreamWrap extends InputStream {
	private InputStream in;
	private String enc;
	private boolean reverse;

	public InputStreamWrap(InputStream in, String enc) {
		this(in, enc, false);
	}

	public InputStreamWrap(InputStream in, String enc, boolean reverse) {
		this.in = in;
		this.enc = enc;
		this.reverse = reverse;
	}

	//----------------------------------------------------------------------

	public int read() throws IOException {
		try {
			return in.read();
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public int read(byte b[]) throws IOException {
		try {
			return in.read(b);
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public int read(byte b[], int off, int len) throws IOException {
		try {
			return in.read(b, off, len);
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public long skip(long n) throws IOException {
		try {
			return in.skip(n);
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public int available() throws IOException {
		try {
			return in.available();
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public void close() throws IOException {
		try {
			in.close();
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public void mark(int readlimit) {
		in.mark(readlimit);
	}

	public void reset() throws IOException {
		try {
			in.reset();
		} catch (IOException e) {
			throw new IOExceptionWrap(e, enc, reverse);
		}
	}

	public boolean markSupported() {
		return in.markSupported();
	}
}

class SQLExceptionWrap extends SQLException {
	private String enc;
	private final static String ASCENC = "8859_1";
	private final static String UNIENC = "UNICODE";

	public SQLExceptionWrap(SQLException e, String enc) {
		super(e.getMessage(), e.getSQLState(), e.getErrorCode());
		super.setNextException(e.getNextException());
		this.enc = enc;
	}
	private String decode(String s) {
		try {
			return (s == null)                     ? null :
			       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
			       (enc.equals(UNIENC))            ? s :
			       /* else */                        new String(s.getBytes(ASCENC), enc);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	//----------------------------------------------------------------------

	public String getMessage() {
		return decode(super.getMessage());
	}

	public String getSQLState() {
		return decode(super.getSQLState());
	}

	public SQLException getNextException() {
		SQLException e = super.getNextException();
		if (e == null || e instanceof DataTruncationWrap || e instanceof SQLWarningWrap || e instanceof SQLExceptionWrap)
			return e;
		else if (e instanceof DataTruncation)
			return new DataTruncationWrap((DataTruncation)e, enc);
		else if (e instanceof SQLWarning)
			return new SQLWarningWrap((SQLWarning)e, enc);
		else if (e instanceof SQLException)
			return new SQLExceptionWrap(e, enc);
		else
			return e;
	}

	public void setNextException(SQLException e) {
		if (e == null || e instanceof DataTruncationWrap || e instanceof SQLWarningWrap || e instanceof SQLExceptionWrap)
			super.setNextException(e);
		else if (e instanceof DataTruncation)
			super.setNextException(new DataTruncationWrap((DataTruncation)e, UNIENC));
		else if (e instanceof SQLWarning)
			super.setNextException(new SQLWarningWrap((SQLWarning)e, UNIENC));
		else if (e instanceof SQLException)
			super.setNextException(new SQLExceptionWrap(e, UNIENC));
		else
			super.setNextException(e);
	}
}

class SQLWarningWrap extends SQLWarning {
	private String enc;
	private final static String ASCENC = "8859_1";
	private final static String UNIENC = "UNICODE";

	public SQLWarningWrap(SQLWarning w, String enc) {
		super(w.getMessage(), w.getSQLState(), w.getErrorCode());
		super.setNextException(w.getNextException());
		super.setNextWarning(w.getNextWarning());
		this.enc = enc;
	}
	private String decode(String s) {
		try {
			return (s == null)                     ? null :
			       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
			       (enc.equals(UNIENC))            ? s :
			       /* else */                        new String(s.getBytes(ASCENC), enc);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	//----------------------------------------------------------------------

	public String getMessage() {
		return decode(super.getMessage());
	}

	public String getSQLState() {
		return decode(super.getSQLState());
	}

	public SQLException getNextException() {
		SQLException e = super.getNextException();
		if (e == null || e instanceof DataTruncationWrap || e instanceof SQLWarningWrap || e instanceof SQLExceptionWrap)
			return e;
		else if (e instanceof DataTruncation)
			return new DataTruncationWrap((DataTruncation)e, enc);
		else if (e instanceof SQLWarning)
			return new SQLWarningWrap((SQLWarning)e, enc);
		else if (e instanceof SQLException)
			return new SQLExceptionWrap(e, enc);
		else
			return e;
	}

	public void setNextException(SQLException e) {
		if (e == null || e instanceof DataTruncationWrap || e instanceof SQLWarningWrap || e instanceof SQLExceptionWrap)
			super.setNextException(e);
		else if (e instanceof DataTruncation)
			super.setNextException(new DataTruncationWrap((DataTruncation)e, UNIENC));
		else if (e instanceof SQLWarning)
			super.setNextException(new SQLWarningWrap((SQLWarning)e, UNIENC));
		else if (e instanceof SQLException)
			super.setNextException(new SQLExceptionWrap(e, UNIENC));
		else
			super.setNextException(e);
	}

	//----------------------------------------------------------------------

	public SQLWarning getNextWarning() {
		SQLWarning w = super.getNextWarning();
		if (w == null || w instanceof DataTruncationWrap || w instanceof SQLWarningWrap)
			return w;
		else if (w instanceof DataTruncation)
			return new DataTruncationWrap((DataTruncation)w, enc);
		else if (w instanceof SQLWarning)
			return new SQLWarningWrap(w, enc);
		else
			return w;
	}

	public void setNextWarning(SQLWarning w) {
		if (w == null || w instanceof DataTruncationWrap || w instanceof SQLWarningWrap)
			super.setNextWarning(w);
		else if (w instanceof DataTruncation)
			super.setNextWarning(new DataTruncationWrap((DataTruncation)w, UNIENC));
		else if (w instanceof SQLWarning)
			super.setNextWarning(new SQLWarningWrap(w, UNIENC));
		else
			super.setNextWarning(w);
	}
}

class DataTruncationWrap extends DataTruncation {
	private String enc;
	private final static String ASCENC = "8859_1";
	private final static String UNIENC = "UNICODE";

	public DataTruncationWrap(DataTruncation t, String enc) {
		super(t.getIndex(), t.getParameter(), t.getRead(), t.getDataSize(), t.getTransferSize());
		super.setNextException(t.getNextException());
		super.setNextWarning(t.getNextWarning());
		this.enc = enc;
	}
	private String decode(String s) {
		try {
			return (s == null)                     ? null :
			       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
			       (enc.equals(UNIENC))            ? s :
			       /* else */                        new String(s.getBytes(ASCENC), enc);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	//----------------------------------------------------------------------

	public String getMessage() {
		return decode(super.getMessage());
	}

	public String getSQLState() {
		return decode(super.getSQLState());
	}

	public SQLException getNextException() {
		SQLException e = super.getNextException();
		if (e == null || e instanceof DataTruncationWrap || e instanceof SQLWarningWrap || e instanceof SQLExceptionWrap)
			return e;
		else if (e instanceof DataTruncation)
			return new DataTruncationWrap((DataTruncation)e, enc);
		else if (e instanceof SQLWarning)
			return new SQLWarningWrap((SQLWarning)e, enc);
		else if (e instanceof SQLException)
			return new SQLExceptionWrap(e, enc);
		else
			return e;
	}

	public void setNextException(SQLException e) {
		if (e == null || e instanceof DataTruncationWrap || e instanceof SQLWarningWrap || e instanceof SQLExceptionWrap)
			super.setNextException(e);
		else if (e instanceof DataTruncation)
			super.setNextException(new DataTruncationWrap((DataTruncation)e, UNIENC));
		else if (e instanceof SQLWarning)
			super.setNextException(new SQLWarningWrap((SQLWarning)e, UNIENC));
		else if (e instanceof SQLException)
			super.setNextException(new SQLExceptionWrap(e, UNIENC));
		else
			super.setNextException(e);
	}

	//----------------------------------------------------------------------

	public SQLWarning getNextWarning() {
		SQLWarning w = super.getNextWarning();
		if (w == null || w instanceof DataTruncationWrap || w instanceof SQLWarningWrap)
			return w;
		else if (w instanceof DataTruncation)
			return new DataTruncationWrap((DataTruncation)w, enc);
		else if (w instanceof SQLWarning)
			return new SQLWarningWrap(w, enc);
		else
			return w;
	}

	public void setNextWarning(SQLWarning w) {
		if (w == null || w instanceof DataTruncationWrap || w instanceof SQLWarningWrap)
			super.setNextWarning(w);
		else if (w instanceof DataTruncation)
			super.setNextWarning(new DataTruncationWrap((DataTruncation)w, UNIENC));
		else if (w instanceof SQLWarning)
			super.setNextWarning(new SQLWarningWrap(w, UNIENC));
		else
			super.setNextWarning(w);
	}
}

class IOExceptionWrap extends IOException {
	private String enc;
	private boolean reverse;
	private final static String ASCENC = "8859_1";

	public IOExceptionWrap(IOException e, String enc) {
		this(e, enc, false);
	}

	public IOExceptionWrap(IOException e, String enc, boolean reverse) {
		super(e.getMessage());
		this.enc = enc;
		this.reverse = reverse;
	}

	private String decode(String s) {
		try {
			return (s == null)                     ? null :
			       (enc == null || enc.equals("")) ? new String(s.getBytes(ASCENC)) :
			       /* else */                        new String(s.getBytes(ASCENC), enc);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	private String encode(String s) {
		try {
			return (s == null)                     ? null :
			       (enc == null || enc.equals("")) ? new String(s.getBytes(), ASCENC) :
			       /* else */                        new String(s.getBytes(enc), ASCENC);
		} catch (UnsupportedEncodingException e) {
			return null;
		}
	}

	//----------------------------------------------------------------------

	public String getMessage() {
		String s = super.getMessage();
		return (reverse) ? encode(s) : decode(s);
	}
}
